"""
Global definitions for the library.
"""

import operator

OPS = {
    ">": operator.gt,
    "<": operator.lt,
    ">=": operator.ge,
    "<=": operator.le,
    "==": operator.eq,
}

PLOTLYSIZELIMIT = 300000
